#pragma once

#include <lektor/common.h>
#include <lektor/mkv.h>
#include <lektor/uri.h>

#include <sqlite3.h>
#include <mpv/client.h>

#include <stdbool.h>
#include <linux/limits.h>

struct lkt_queue_state {
    int volume;
    int paused;
    int random;
    int repeat;
    int single;
    int current;
    int duration;
    int consume;
    int length;
};

long database_get_timestamp(volatile sqlite3 *db);
void database_stamp        (volatile sqlite3 *db);
void database_updated      (volatile sqlite3 *db);

/* Open correctly a database for lektor. */
bool database_new   (volatile sqlite3 **db);
bool database_init  (const char *dbpath);
bool database_open  (volatile sqlite3 *db, const char *dbpath);
bool database_attach(volatile sqlite3 *db, const char *name, const char *dbpath);
bool database_detach(volatile sqlite3 *db, const char *name);

/* Get information on the queue and currently playing kara. */
bool database_queue_state       (volatile sqlite3 *db, struct lkt_queue_state *res);
bool database_queue_current_kara(volatile sqlite3 *db, struct kara_metadata *res, int *id);

bool database_queue_set_paused       (volatile sqlite3 *db, bool paused);
bool database_queue_set_current_index(volatile sqlite3 *db, int idx);
bool database_queue_get_current_file (volatile sqlite3 *db, char filepath[PATH_MAX]);

/* Mpv and database synchronisation. */
bool database_sync_mpv_state(volatile sqlite3 *db, mpv_handle **mpv);

/* Update the database. */
bool database_update              (volatile sqlite3 *db, const char *kara_dir);
bool database_update_add          (volatile sqlite3 *db, const char *kara_path, struct kara_metadata *mdt, uint64_t id, bool avail);
bool database_update_set_available(volatile sqlite3 *db, uint64_t id);

/* Control the content of the queue. */
bool database_queue_add_id      (volatile sqlite3 *db, int id,               int priority);
bool database_queue_add_query   (volatile sqlite3 *db, const char *query,    int priority);
bool database_queue_add_author  (volatile sqlite3 *db, const char *author,   int priority);
bool database_queue_add_language(volatile sqlite3 *db, const char *author,   int priority);
bool database_queue_add_category(volatile sqlite3 *db, const char *author,   int priority);
bool database_queue_add_type    (volatile sqlite3 *db, const char *author,   int priority);
bool database_queue_add_plt     (volatile sqlite3 *db, const char *plt_name, int priority);

bool database_queue_insert_query   (volatile sqlite3 *db, const char *query, int pos);
bool database_queue_insert_id      (volatile sqlite3 *db, int id);
bool database_queue_insert_author  (volatile sqlite3 *db, const char *author);
bool database_queue_insert_language(volatile sqlite3 *db, const char *author);
bool database_queue_insert_category(volatile sqlite3 *db, const char *author);
bool database_queue_insert_type    (volatile sqlite3 *db, const char *author);
bool database_queue_insert_plt     (volatile sqlite3 *db, const char *plt_name);

bool database_queue_del_id (volatile sqlite3 *db, int id);
bool database_queue_del_pos(volatile sqlite3 *db, int pos);
bool database_queue_clear  (volatile sqlite3 *db);
bool database_queue_crop   (volatile sqlite3 *db);
bool database_queue_move   (volatile sqlite3 *db, int from, int to);
bool database_queue_shuffle(volatile sqlite3 *db);
bool database_queue_seekid (volatile sqlite3 *db, int id, int *out_pos);

/* Control the playing state of the queue. */
bool database_queue_toggle_pause(volatile sqlite3 *db);
bool database_queue_play        (volatile sqlite3 *db, int pos);
bool database_queue_stop        (volatile sqlite3 *db);

/* A search callback to be called after each matched row */
struct lkt_callback {
    bool (*call)(void *args, int id, int id_len, const char *sql_row);
    struct lkt_state *srv;
    size_t c;
    int iterations;
};

/* List the content of the queue */
bool database_queue_list(volatile sqlite3 *db, size_t from, size_t to, struct lkt_callback *callback);

/* Search the database */
struct lkt_search {
    volatile sqlite3 *db;
    sqlite3_stmt *stmt;
    enum lkt_search_type {
        lkt_search_database,
        lkt_search_playlist,
        lkt_search_queue,
    } type;
    void (*call)(void); /* Will be casted. */
    struct lkt_state *srv;
    size_t c;
    long continuation;  /* Is this a continuation from a previous command? */
    int msg_count;      /* How much messages we can send. */
};

typedef bool (*lkt_search_database_func)(struct lkt_state *srv, size_t c, int id, int id_len,
        const char *row);

bool database_search_database_init(volatile sqlite3 *db, char *col_name, char *rgx, struct lkt_search *ret);
bool database_search_queue_init   (volatile sqlite3 *db, char *col_name, char *rgx, struct lkt_search *ret);
bool database_search_iter(struct lkt_search *item);

/* Next and prev operation on the queue. */
bool database_queue_next(volatile sqlite3 *db, char filepath[PATH_MAX]);
bool database_queue_prev(volatile sqlite3 *db, char filepath[PATH_MAX]);

/* Set a value in the config table */
bool database_config_set     (volatile sqlite3 *db, const char *section, const char *key, const char *value);
bool database_config_get_text(volatile sqlite3 *db, const char *section, const char *key, char *value, size_t len);
bool database_config_get_int (volatile sqlite3 *db, const char *section, const char *key, int *value);
bool database_config_exists  (volatile sqlite3 *db, const char *section, const char *key);
bool database_config_queue   (volatile sqlite3 *db, const char *option,  int value);
bool database_config_queue_default(volatile sqlite3 *db);

/* Get a value from the config table */
bool database_get_config(volatile sqlite3 *db, const char *option, int *value);

/* Playlists */
struct lkt_playlist_metadata {
    const char *name;
    int last_update;
    int content_count;
};

/* Control playlists */
bool database_plt_create    (volatile sqlite3 *db, const char *name);
bool database_plt_remove    (volatile sqlite3 *db, const char *name);
bool database_plt_remove_pos(volatile sqlite3 *db, const char *name, int pos);
bool database_plt_clear     (volatile sqlite3 *db, const char *name);
bool database_plt_rename    (volatile sqlite3 *db, const char *old_name, const char *new_name);

bool database_plt_export(volatile sqlite3 *db, const char *name);
bool database_plt_import(volatile sqlite3 *db, const char *name);

bool database_plt_add_uri(volatile sqlite3 *db, const char *name, struct lkt_uri *uri);

/* User control, yeah, MPD authentification sucks. */
bool database_user_authentificate(volatile sqlite3 *db, const char *password);
bool database_user_add(volatile sqlite3 *db, const char *username, const char *password);

/* Stickers manipulations. */
struct sticker_callback {
    struct lkt_state *srv;
    size_t c;
    int uri, value, is_ok;
    const char *name;
    bool (*call)(void *args, const char *sticker, const char *type, int uri, int value);
};

bool database_sticker_create        (volatile sqlite3 *db, const char *name);
bool database_sticker_delete        (volatile sqlite3 *db, const char *name);
bool database_sticker_delete_specify(volatile sqlite3 *sb, const char *type, int uri, const char *name);
bool database_sticker_list          (volatile sqlite3 *db, const char *type, struct sticker_callback *call);
bool database_sticker_set           (volatile sqlite3 *db, const char *type, const char *name, int uri, int value);
bool database_sticker_get           (volatile sqlite3 *db, const char *type, const char *name, int uri, struct sticker_callback *call);
