#define _POSIX_C_SOURCE 200809L

#include <lektor/database.h>
#include <lektor/common.h>

#include <linux/limits.h>
#include <stdio.h>
#include <string.h>

/* Find in in database/open.c */
extern int is_sql_str_invalid(const char *);

bool
database_queue_state(volatile sqlite3 *db, struct lkt_queue_state *res)
{
    static const char *SQL_STMT =
        "SELECT"
        "  volume, paused, random, repeat, single, current, duration, consume, "
        "  (SELECT COUNT(*) FROM queue) AS length "
        "FROM queue_state;\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_STEP_ROW(db, stmt, error);

    res->volume = sqlite3_column_int(stmt, 0);
    res->paused = sqlite3_column_int(stmt, 1);
    res->random = sqlite3_column_int(stmt, 2);
    res->repeat = sqlite3_column_int(stmt, 3);
    res->single = sqlite3_column_int(stmt, 4);

    if (sqlite3_column_type(stmt, 5) == SQLITE_NULL)
        res->current = -1;
    else
        res->current = sqlite3_column_int(stmt, 5);

    if (sqlite3_column_type(stmt, 6) == SQLITE_NULL)
        res->duration = 0;
    else
        res->duration = sqlite3_column_int(stmt, 6);

    res->consume = sqlite3_column_int(stmt, 7);
    res->length = sqlite3_column_int(stmt, 8);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_queue_current_kara(volatile sqlite3 *db, struct kara_metadata *res, int *id)
{
    static const char *SQL_STMT =
        "SELECT song_name, source_name, category, language, author_name, song_type, song_number, kara_id"
        " FROM kara"
        " JOIN queue_ ON kara_id = kara.id"
        " JOIN queue_state ON current = position";
    sqlite3_stmt *stmt = 0;
    int ret = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_STEP_ROW(db, stmt, error);

    /* Here use gotos because of optimisations done by compilators.
       Most of the time it won't be NULL. */
    if (!res)
        goto no_metadata;
    strncpy(res->song_name,   (const char *) sqlite3_column_text(stmt, 0), LEKTOR_TAG_MAX - 1);
    strncpy(res->source_name, (const char *) sqlite3_column_text(stmt, 1), LEKTOR_TAG_MAX - 1);
    strncpy(res->category,    (const char *) sqlite3_column_text(stmt, 2), LEKTOR_TAG_MAX - 1);
    strncpy(res->language,    (const char *) sqlite3_column_text(stmt, 3), LEKTOR_TAG_MAX - 1);
    strncpy(res->author_name, (const char *) sqlite3_column_text(stmt, 4), LEKTOR_TAG_MAX - 1);
    strncpy(res->song_type,   (const char *) sqlite3_column_text(stmt, 5), LEKTOR_TAG_MAX - 1);
    res->song_number = sqlite3_column_int(stmt, 6);
no_metadata:
    /* Most of the time this will be NULL. */
    if (id && sqlite3_column_type(stmt, 7) != SQLITE_NULL)
        *id = sqlite3_column_int(stmt, 7);

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

static bool
queue_add_with_col_like_str(volatile sqlite3 *db, const char *col, const char *val, int priority)
{
    RETURN_IF(is_sql_str_invalid(col), "Column name is invalid", false);

    char SQL[LKT_MAX_SQLITE_STATEMENT];
    static const char *SQL_STMT =
        "INSERT INTO queue (kara_id, priority)"
        "  SELECT id, ?"
        "  FROM kara"
        "  WHERE %s LIKE ? AND available = 1"
        "  ORDER BY RANDOM();";
    bool status = false;
    sqlite3_stmt *stmt = NULL;

    snprintf(SQL, LKT_MAX_SQLITE_STATEMENT - 1, SQL_STMT, col);
    SQL[LKT_MAX_SQLITE_STATEMENT - 1] = 0;
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, priority, error);
    SQLITE_BIND_TEXT(db, stmt, 2, val, error);
    SQLITE_STEP_DONE(db, stmt, error);
    status = true;
error:
    sqlite3_finalize(stmt);
    return status;
}

static bool
queue_insert_with_col_like_str(volatile sqlite3 *db, const char *col, const char *val, int pos)
{
    RETURN_IF(is_sql_str_invalid(col), "Column name is invalid", false);

    char SQL[LKT_MAX_SQLITE_STATEMENT];
    static const char *SQL_STMT =
        "BEGIN TRANSACTION;"
        "CREATE TEMPORARY TABLE queue_tmp"
        "  ( position_tmp INTEGER PRIMARY KEY AUTOINCREMENT"
        "  , kara_id_tmp INTEGER"
        "  , priority_tmp INTEGER NOT NULL DEFAULT 1 CHECK(priority > 0 AND priority < 6)"
        "  );"
        "INERT INTO queue_tmp (kara_id_tmp, priority_tmp)"
        "  SELECT (kara_id, priority)"
        "  FROM queue"
        "  WHERE position >= ?"
        "  ORDER BY position ASC;"
        "INSERT INTO queue (kara_id, priority)"
        "SELECT id, priority"
        "  FROM kara"
        "  WHERE %s LIKE ? AND available = 1"
        "  ORDER BY RANDOM();"
        "INSERT INTO queue (kara_id, priority)"
        "  SELECT kara_id_tmp, priority"
        "  FROM queue_tmp"
        "  ORDER BY position ASC;"
        "DROP TABLE queue_tmp;"
        "COMMIT TRANSACTION;";
    bool status = false;
    sqlite3_stmt *stmt = NULL;

    snprintf(SQL, LKT_MAX_SQLITE_STATEMENT - 1, SQL_STMT, col);
    SQL[LKT_MAX_SQLITE_STATEMENT - 1] = 0;
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, pos, error);
    SQLITE_BIND_TEXT(db, stmt, 2, val, error);
    SQLITE_STEP_DONE(db, stmt, error);
    status = true;
error:
    sqlite3_finalize(stmt);
    return status;
}

bool
database_queue_insert_query(volatile sqlite3 *db, const char *query, int pos)
{
    bool status = queue_insert_with_col_like_str(db, LKT_DATABASE_KARA_ALL, query, pos);
    LOG_INFO("%s '%s' at %d", status ? "Successfull add query" : "Failed add query", query, pos);
    return status;
}

bool
database_queue_add_plt(volatile sqlite3 *db, const char *plt_name, int priority)
{
    static const char *SQL_STMT =
        "INSERT INTO queue (kara_id, priority) "
        "SELECT kara.id, ?"
        "  FROM kara"
        "  JOIN kara_playlist ON kara_id = kara.id AND kara.available = 1"
        "  JOIN playlist ON playlist_id = playlist.id AND playlist.name = ?"
        "  ORDER BY RANDOM();";
    bool status = false;
    sqlite3_stmt *stmt = NULL;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, priority, error);
    SQLITE_BIND_TEXT(db, stmt, 2, plt_name, error);
    SQLITE_STEP_DONE(db, stmt, error);
    status = true;
error:
    sqlite3_finalize(stmt);
    return status;
}

bool
database_queue_add_query(volatile sqlite3 *db, const char *query, int priority)
{
    bool status = queue_add_with_col_like_str(db, LKT_DATABASE_KARA_ALL, query, priority);
    LOG_INFO("%s '%s' with prio %d", status ? "Successfull add query" : "Failed add query", query, priority);
    return status;
}

bool
database_queue_add_author(volatile sqlite3 *db, const char *author, int priority)
{
    bool status = queue_add_with_col_like_str(db, LKT_DATABASE_NAME_KAUTHOR, author, priority);
    LOG_INFO("%s '%s'", status ? "Successfull add author" : "Failed add author", author);
    return status;
}

bool
database_queue_add_language(volatile sqlite3 *db, const char *language, int priority)
{
    bool status = queue_add_with_col_like_str(db, LKT_DATABASE_NAME_KLANG, language, priority);
    LOG_INFO("%s '%s'", status ? "Successfull add language" : "Failed add language", language);
    return status;
}

bool
database_queue_add_category(volatile sqlite3 *db, const char *cat, int priority)
{
    bool status = queue_add_with_col_like_str(db, LKT_DATABASE_NAME_KCAT, cat, priority);
    LOG_INFO("%s '%s'", status ? "Successfull add category" : "Failed add category", cat);
    return status;
}

bool
database_queue_add_type(volatile sqlite3 *db, const char *type, int priority)
{
    bool status = queue_add_with_col_like_str(db, LKT_DATABASE_NAME_KTYPE, type, priority);
    LOG_INFO("%s '%s'", status ? "Successfull add type" : "Failed add type", type);
    return status;
}

bool
database_queue_add_id(volatile sqlite3 *db, int id, int priority)
{
    static const char *SQL_STMT = "INSERT INTO queue (kara_id, priority) VALUES (?,?);";
    bool status = false;
    sqlite3_stmt *stmt = NULL;
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, id, error);
    SQLITE_BIND_INT(db, stmt, 2, priority, error);
    SQLITE_STEP_DONE(db, stmt, error);
    status = true;
error:
    sqlite3_finalize(stmt);
    return status;
}

bool
database_queue_del_id(volatile sqlite3 *db, int id)
{
    static const char *SQL_TEMPLATE =
        "BEGIN TRANSACTION;"
        "WITH before(pos) AS (SELECT position FROM queue_ JOIN queue_state WHERE position < current ORDER BY position DESC LIMIT 1) "
        "UPDATE queue_state SET current = CASE"
        " WHEN current IS NULL THEN NULL"
        " ELSE (SELECT pos FROM before) "
        "END WHERE current > (SELECT position FROM queue_ WHERE kara_id = %d);"
        "DELETE FROM queue WHERE kara_id = %d;"
        "COMMIT TRANSACTION;";
    char SQL[LKT_MAX_SQLITE_STATEMENT];
    snprintf(SQL, LKT_MAX_SQLITE_STATEMENT - 1, SQL_TEMPLATE, id, id, id);
    SQL[LKT_MAX_SQLITE_STATEMENT - 1] = '\0';
    SQLITE_EXEC(db, SQL, error);
    return true;
error:
    return false;
}

bool
database_queue_next(volatile sqlite3 *db, char filepath[PATH_MAX])
{
    static const char *SQL_STMT =
        "SELECT file_path, position, RANDOM()"
        " FROM kara"
        " JOIN queue_ ON kara.id = queue_.kara_id"
        " JOIN queue_state"
        "  ON available = 1 AND current IS NULL OR"
        "  CASE WHEN (SELECT single FROM queue_state) = 1 THEN position = current"
        "       WHEN (SELECT random FROM queue_state) = 1 THEN position <> current"
        "       ELSE position > current END"
        " ORDER BY CASE WHEN (SELECT random FROM queue_state) = 1 THEN RANDOM() ELSE 2 END"
        " LIMIT 1";
    char SQL_UPDATE[LKT_MAX_SQLITE_STATEMENT];
    bool status = false;
    int code = SQLITE_OK, id;
    sqlite3_stmt *stmt = NULL;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);

    code = sqlite3_step(stmt);

    if (code == SQLITE_ROW) {
        id = MAX(1, sqlite3_column_int(stmt, 1));
        snprintf(SQL_UPDATE, LKT_MAX_SQLITE_STATEMENT - 1, "UPDATE queue_state SET current = %d;", id);
        SQL_UPDATE[LKT_MAX_SQLITE_STATEMENT - 1] = 0;

        if (filepath != NULL)
            strncpy(filepath, (const char *) sqlite3_column_text(stmt, 0), PATH_MAX);
        else {
            LOG_ERROR("Failed to get file, id was %d", id);
            goto error;
        }
    }

    else if (code == SQLITE_DONE) {
        LOG_ERROR("%s", "Failed to get next, don't set current to NULL");
        goto error;
    }

    else {
        LOG_ERROR_SCT("DB", "Failed to fetch next kara: %s", sqlite3_errmsg((sqlite3 *) db));
        goto error;
    }

    SQLITE_EXEC(db, SQL_UPDATE, error);
    status = true;
error:
    sqlite3_finalize(stmt);
    return status;
}

bool
database_queue_toggle_pause(volatile sqlite3 *db)
{
    static const char *SQL_STMT = "UPDATE queue_state SET paused = 1 - paused";
    SQLITE_EXEC(db, SQL_STMT, error);
    return true;
error:
    return false;
}

bool
database_queue_prev(volatile sqlite3 *db, char filepath[PATH_MAX])
{
    static const char *SQL_STMT =
        "SELECT file_path, position "
        "  FROM kara"
        "  JOIN queue_ ON kara.id = queue_.kara_id AND available = 1"
        "  JOIN queue_state ON CASE "
        "    WHEN (SELECT single FROM queue_state LIMIT 1) = 1 THEN position = current"
        "    ELSE queue_.position < queue_state.current"
        "  END"
        "  ORDER BY position DESC LIMIT 1;";
    char SQL_UPDATE[LKT_MAX_SQLITE_STATEMENT];
    bool status = false;
    int code = SQLITE_OK, id;
    sqlite3_stmt *stmt = NULL;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    code = sqlite3_step(stmt);

    if (code == SQLITE_ROW) {
        id = MAX(1, sqlite3_column_int(stmt, 1));
        snprintf(SQL_UPDATE, LKT_MAX_SQLITE_STATEMENT - 1, "UPDATE queue_state SET current = %d;", id);
        SQL_UPDATE[LKT_MAX_SQLITE_STATEMENT - 1] = 0;

        if (filepath != NULL)
            strncpy(filepath, (const char *) sqlite3_column_text(stmt, 0), PATH_MAX);
        else {
            LOG_ERROR("Failed to get file, position was %d", id);
            goto error;
        }
    }

    else if (code == SQLITE_DONE) {
        LOG_ERROR("%s", "Failed to get previous, don't set current to NULL");
        goto error;
    }

    else {
        LOG_ERROR_SCT("DB", "Failed to fetch prev kara: %s", sqlite3_errmsg((sqlite3 *) db));
        goto error;
    }

    SQLITE_EXEC(db, SQL_UPDATE, error);
    status = true;
error:
    sqlite3_finalize(stmt);
    return status;
}

bool
database_queue_clear(volatile sqlite3 *db)
{
    static const char *SQL_STMT =
        "BEGIN TRANSACTION;"
        "DELETE FROM queue;"
        "DELETE FROM sqlite_sequence WHERE name = 'queue';"
        "UPDATE queue_state SET current = NULL;"
        "COMMIT TRANSACTION;";
    SQLITE_EXEC(db, SQL_STMT, error);
    return true;
error:
    return false;
}

bool
database_queue_crop(volatile sqlite3 *db)
{
    static const char *SQL_STMT =
        "DELETE FROM queue WHERE queue.kara_id <> (SELECT current FROM queue_state LIMIT 1);";
    SQLITE_EXEC(db, SQL_STMT, error);
    return true;
error:
    return false;
}

bool
database_queue_move(volatile sqlite3 *db, int from, int to)
{
    /* stmt arguments in order:
     * 1. to
     * 2. to
     * 3. from
     */
    static const char *SQL_STMT =
        "BEGIN TRANSACTION;"
        "UPDATE queue SET position = position + 1 WHERE position >= ?;"
        "UPDATE queue SET position = ? WHERE position = ?;"
        "COMMIT TRANSACTION;";
    sqlite3_stmt *stmt = 0;
    bool ret = false;
    int code;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, to, error);
    SQLITE_BIND_INT(db, stmt, 2, to, error);
    SQLITE_BIND_INT(db, stmt, 3, from, error);

    code = sqlite3_step(stmt);

    if (code != SQLITE_ROW && code != SQLITE_DONE && code != SQLITE_OK) {
        LOG_ERROR_SCT("DB", "%s", "Move failed");
        goto error;
    }

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_queue_play(volatile sqlite3 *db, int pos)
{
    static const char *SQL_STMT =
        "UPDATE queue_state SET current = ?, paused = 0"
        "  WHERE current <= 0 OR current IS NULL";
    int code = SQLITE_OK;
    bool ret = false;
    sqlite3_stmt *stmt = NULL;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, pos, error);

    code = sqlite3_step(stmt);

    if (code != SQLITE_OK && code != SQLITE_DONE) {
        LOG_ERROR_SCT("DB", "Failed to update queue_state: %s", sqlite3_errmsg((sqlite3 *) db));
        goto error;
    }

    ret = true;
error:
    return ret;
}

bool
database_queue_stop(volatile sqlite3 *db)
{
    static const char *SQL_STMT = "UPDATE queue_state SET current = NULL;";
    SQLITE_EXEC(db, SQL_STMT, error);
    return true;
error:
    return false;
}

bool
database_queue_set_current_index(volatile sqlite3 *db, int idx)
{
    static const char *SQL_GET_TEMPLATE = "UPDATE queue_state SET current = %d;";
    char SQL_GET[LKT_MAX_SQLITE_STATEMENT];

    if (idx <= 0) {
        LOG_ERROR_SCT("DB", "An idx of %d is invalid, must be >= 0", idx);
        return false;
    }

    RETURN_IF(snprintf(SQL_GET, LKT_MAX_SQLITE_STATEMENT - 1, SQL_GET_TEMPLATE, idx) < 0, "Snprintf failed",
              false);
    SQL_GET[LKT_MAX_SQLITE_STATEMENT - 1] = 0;
    SQLITE_EXEC(db, SQL_GET, error);
    return true;
error:
    return false;
}

bool
database_queue_get_current_file(volatile sqlite3 *db, char filepath[PATH_MAX])
{
    static const char *SQL_STMT =
        "SELECT file_path"
        "  FROM kara"
        "  JOIN queue_ ON kara.id = queue_.kara_id"
        "  JOIN queue_state ON queue_.position = queue_state.current";
    bool status = false;
    int code = SQLITE_OK;
    sqlite3_stmt *stmt = NULL;

    if (filepath == NULL)
        goto error;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    code = sqlite3_step(stmt);

    if (code == SQLITE_ROW)
        strncpy(filepath, (const char *) sqlite3_column_text(stmt, 0), PATH_MAX);

    else {
        LOG_ERROR_SCT("DB", "Failed to fetch prev kara: %s", sqlite3_errmsg((sqlite3 *) db));
        goto error;
    }

    status = true;
error:
    sqlite3_finalize(stmt);
    return status;
}

bool
database_queue_set_paused(volatile sqlite3 *db, bool paused)
{
    const char *SQL = paused ? "UPDATE queue_state SET paused = 1;" : "UPDATE queue_state SET paused = 0;";
    SQLITE_EXEC(db, SQL, error);
    return true;
error:
    return false;
}

bool
database_queue_shuffle(volatile sqlite3 *db)
{
    const char *SQL =
        "BEGIN TRANSACTION;"
        /* Create temporary queue.  */
        "CREATE TEMPORARY TABLE queue_tmp"
        "  ( position INTEGER PRIMARY KEY AUTOINCREMENT CHECK(position > 0)"
        "  , kara_id INTEGER"
        "  , priority INTEGER NOT NULL DEFAULT 1 CHECK(priority > 0 AND priority < 6)"
        "  );"
        /* When current is NULL, that thing is also NULL, so no insertion is done.  */
        "INSERT INTO queue_tmp (kara_id, priority)"
        "  SELECT kara_id, priority"
        "  FROM queue"
        "  JOIN queue_state ON queue.position = queue_state.current;"
        /* Insert Max priority in order.  */
        "INSERT INTO queue_tmp (kara_id, priority)"
        "  SELECT kara_id, priority"
        "  FROM queue"
        "  JOIN queue_state ON"
        "    queue.position <> CASE"
        "      WHEN queue_state.current IS NOT NULL THEN queue_state.current"
        "      ELSE 0 END"
        "    AND queue.priority = 5;"
        /* Insert the others karas, where the priority is < 5.  */
        "INSERT INTO queue_tmp (kara_id, priority)"
        "  SELECT kara_id, priority"
        "  FROM queue"
        "  JOIN queue_state ON"
        "    queue.position <> CASE"
        "      WHEN queue_state.current IS NOT NULL THEN queue_state.current"
        "      ELSE 0 END"
        "    AND queue.priority <> 5"
        "  ORDER BY RANDOM();"
        /* Do the insertion in the other way.  */
        "DELETE FROM queue;"
        "DELETE FROM sqlite_sequence WHERE name = 'queue';"
        "INSERT INTO queue (position, kara_id, priority)"
        "  SELECT position, kara_id, priority"
        "  FROM queue_tmp"
        "  ORDER BY priority, position;"
        "DROP TABLE queue_tmp;"
        "DELETE FROM sqlite_sequence WHERE name = 'queue_tmp';"
        /* Set the current to the right kara when needed.  */
        "UPDATE queue_state"
        "  SET current = CASE"
        "    WHEN current NOT NULL THEN 1"
        "    ELSE NULL END;"
        "COMMIT;";
    SQLITE_EXEC(db, SQL, error);
    return true;
error:
    return false;
}

bool
database_queue_list(volatile sqlite3 *db, size_t from, size_t to, struct lkt_callback *callback)
{
    const char *SQL_STMT =
        "WITH content AS ("
        " SELECT kara.id AS id, string, LENGTH(CAST(kara.id AS TEXT)) AS len"
        "  FROM queue_"
        "  JOIN kara ON kara_id = kara.id"
        "  WHERE position >= ? AND position <= ?"
        "  GROUP BY position ORDER BY position)"
        "SELECT id, string, (SELECT MAX(len) FROM content) FROM content;";
    int code, id, id_len;
    const char *row;
    bool ret = false;
    sqlite3_stmt *stmt;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, (int) from, error);
    SQLITE_BIND_INT(db, stmt, 2, (int) to, error);

    for (;;) {
        code = sqlite3_step(stmt);


        if (code == SQLITE_ROW) {
            id = sqlite3_column_int(stmt, 0);
            row = (const char *) sqlite3_column_text(stmt, 1);
            id_len = sqlite3_column_int(stmt, 2);
            ++(callback->iterations);
            if (callback->call(callback, id, id_len, row))
                continue;
            else
                break;
        }

        else if (code == SQLITE_OK || code == SQLITE_DONE)
            goto done;

        else
            break;
    }

done:
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_queue_seekid(volatile sqlite3 *db, int id, int *out_pos)
{
    static const char *SQL_STMT = "SELECT position FROM queue_ WHERE kara_id = ? LIMIT 1";
    int ret = 0;
    sqlite3_stmt *stmt;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, id, error);
    SQLITE_STEP_ROW(db, stmt, error);
    *out_pos = sqlite3_column_int(stmt, 0);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}
