#define _POSIX_C_SOURCE 200809L

#include <lektor/common.h>
#include <lektor/database.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <stdio.h>

bool
database_sticker_create(volatile sqlite3 *db, const char *name)
{
    static const char *INSERT =
        "WITH id_max (id) AS (SELECT MAX(id) FROM 'stickers')"
        "INSERT INTO 'stickers' (id, name)"
        " SELECT CASE WHEN id_max.id IS NULL THEN 1 ELSE id_max.id + 1 END, ?"
        " FROM id_max"
        "WHERE NOT EXISTS (SELECT 1 FROM 'stickers' WHERE name = ?);";
    sqlite3_stmt *stmt;
    int ret = false;
    RETURN_IF(strlen(name) == 0, "A sticker name must be at least one character long", ret);
    SQLITE_PREPARE(db, stmt, INSERT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);
    SQLITE_BIND_TEXT(db, stmt, 2, name, error);
    SQLITE_STEP_OK(db, stmt, error);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_sticker_delete(volatile sqlite3 *db, const char *name)
{
    static const char *INSERT = "DELETE FROM 'stickers' WHERE name = ?;";
    sqlite3_stmt *stmt;
    int ret = false;
    RETURN_IF(strlen(name) == 0, "A sticker name must be at least one character long", ret);
    SQLITE_PREPARE(db, stmt, INSERT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);
    SQLITE_STEP_OK(db, stmt, error);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_sticker_list(volatile sqlite3 *db, const char *type, struct sticker_callback *call)
{
    const char *SQL = NULL;
    int ret = false, uri, value;
    sqlite3_stmt *stmt;
    const char *sticker;

    if (type == NULL)
        SQL =
            "SELECT name, sts.id, value FROM 'stickers' LEFT OUTER JOIN "
            "( SELECT id, sticker, value FROM 'stickers.song'"
            "  UNION"
            "  SELECT id, sticker, value FROM 'stickers.plt'"
            ") AS sts"
            "ON sts.sticker = 'stickers'.id;";
    else if (!strcasecmp(type, "plt"))
        SQL =
            "SELECT name, 'stickers.plt'.id, value "
            "FROM 'stickers' "
            "LEFT OUTER JOIN 'stickers' "
            "ON 'stickers'.id = 'stickers.plt'.sticker;";
    else if (!strcasecmp(type, "song"))
        SQL =
            "SELECT name, 'stickers.song'.id, value "
            "FROM 'stickers' "
            "LEFT OUTER JOIN 'stickers' "
            "ON 'stickers'.id = 'stickers.song'.sticker;";
    else {
        LOG_ERROR_SCT("DB", "Type '%s' is invalid", type);
        return false;
    }

    SQLITE_PREPARE(db, stmt, SQL, error);

    for (;;) {
        switch (sqlite3_step(stmt)) {
        case SQLITE_ROW:
            sticker = (const char *) sqlite3_column_text(stmt, 0);
            uri     = sqlite3_column_int(stmt, 1);
            value   = sqlite3_column_int(stmt, 2);
            if (!call->call(call, sticker, type, uri, value))
                goto end_loop;
            continue;

        case SQLITE_DONE:
            goto end_loop;

        default:
            goto error;
        }
    }

end_loop:
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_sticker_set(volatile sqlite3 *db, const char *type, const char *name, int uri, int value)
{
    const char *SQL = NULL;
    sqlite3_stmt *stmt;
    int ret = false;

    /* Bindings:
     * 1 -> the uri of the plalist or kara
     * 2 -> the value of the sticker
     * 3 -> the name of the sticker
     */

    if (!strcasecmp(type, "song"))
        SQL =
            "INSERT OR REPLACE INTO 'stickers.song' (id, sticker, value) "
            "SELECT ?, 'stickers'.id, ? "
            "FROM 'stickers'"
            "WHERE 'stickers'.id = ?;\n";
    else if (!strcasecmp(type, "plt"))
        SQL =
            "INSERT OR REPLACE INTO 'stickers.plt' (id, sticker, value) "
            "SELECT ?, 'stickers'.id, ? "
            "FROM 'stickers'"
            "WHERE 'stickers'.id = ?;\n";
    else {
        LOG_ERROR_SCT("DB", "Type '%s' is invalid", type);
        return false;
    }

    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, uri, error);
    SQLITE_BIND_INT(db, stmt, 2, value, error);
    SQLITE_BIND_TEXT(db, stmt, 3, name, error);
    SQLITE_STEP_OK(db, stmt, error);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_sticker_get(volatile sqlite3 *db, const char *type, const char *name, int uri, struct sticker_callback *call)
{
    const char *SQL = NULL;
    sqlite3_stmt *stmt;
    int ret = false, value;

    if (!strcasecmp(type, "song"))
        SQL =
            "SELECT value "
            "FROM 'stickers' "
            "JOIN 'stickers.song'"
            " ON 'stickers'.id = 'stickers.song'.sticker"
            " AND 'stickers'.name = ?;\n";
    else if (!strcasecmp(type, "plt"))
        SQL =
            "SELECT value "
            "FROM 'stickers' "
            "JOIN 'stickers.plt'"
            " ON 'stickers'.id = 'stickers.plt'.sticker"
            " AND 'stickers'.name = ?;\n";
    else {
        LOG_ERROR_SCT("DB", "Type '%s' is invalid", type);
        return false;
    }

    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);

    for (;;) {
        switch (sqlite3_step(stmt)) {
        case SQLITE_ROW:
            value = sqlite3_column_int(stmt, 1);
            if (!call->call(call, name, type, uri, value))
                goto end_loop;
            continue;

        case SQLITE_DONE:
            goto end_loop;

        default:
            goto error;
        }
    }

end_loop:
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_sticker_delete_specify(volatile sqlite3 *db, const char *type, int uri, const char *name)
{
    char SQL[LKT_MAX_SQLITE_STATEMENT];
    sqlite3_stmt *stmt;
    int ret = false;

    /* Base query. */
    if (strcasecmp("plt", type) && strcasecmp("song", type)) {
        LOG_ERROR_SCT("DB", "Type '%s' is invalid", type);
        return false;
    }

    snprintf(SQL, LKT_MAX_SQLITE_STATEMENT - 1, "DELETE FROM 'stickers.%s' "
             "WHERE 'stickers.%s' = ? ", type, type);
    SQL[LKT_MAX_SQLITE_STATEMENT - 1] = 0;

    /* If there is a name specified. */
    if (!name) {
        strncat(SQL, ";", LKT_MAX_SQLITE_STATEMENT - 1);
        SQL[LKT_MAX_SQLITE_STATEMENT - 1] = 0;
    } else {
        strncat(SQL, "AND sticker = (SELECT id FROM 'stickers' WHERE name = ? LIMIT 1);",
                LKT_MAX_SQLITE_STATEMENT - 1);
        SQL[LKT_MAX_SQLITE_STATEMENT - 1] = 0;
    }

    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, uri, error);

    if (name)
        SQLITE_BIND_TEXT(db, stmt, 2, name, error);

    SQLITE_STEP_OK(db, stmt, error);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}
