#define _POSIX_C_SOURCE 200809L
#define _DEFAULT_SOURCE

#include <common/common.h>
#include <lektor/database.h>
#include <lektor/common.h>

#include <stdbool.h>
#include <sqlite3.h>
#include <stdio.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <linux/limits.h>

static bool
database_add_kara(volatile sqlite3 *db, const char *filename)
{
    RETURN_UNLESS(db || filename, "Invalid argument", false);
    static const char *SQL_STMT =
        "INSERT INTO "
        "kara (song_name, source_name, category, song_type, language, file_path, is_new, author_name, author_year, song_number)"
        "SELECT ?, ?, ?, ?, ?, ?, ?, ?, strftime('%s','now'), ?";
    sqlite3_stmt *stmt = NULL;
    int code = SQLITE_OK, status = false;
    struct kara_metadata data;
    time_t the_time = time(NULL);
    struct tm *the_local_time = localtime(&the_time);
    char year[10];
    code = snprintf(year, 10, "%d", the_local_time->tm_year + 1900);
    RETURN_IF(code < 0 || code >= 10, "Failed to get the year of the current date", false);

    if (kara_metadata_read(&data, filename) != 0) {
        LOG_ERROR("Failed to get mdt for the kara '%s'", filename);
        return false;
    }

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);

    if ((sqlite3_bind_text(stmt, 1, data.song_name,   -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_text(stmt, 2, data.source_name, -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_text(stmt, 3, data.category,    -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_text(stmt, 4, data.song_type,   -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_text(stmt, 5, data.language,    -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_text(stmt, 6, filename,         -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_int (stmt, 7, 0)                       != SQLITE_OK)  ||
        (sqlite3_bind_text(stmt, 8, data.author_name, -1, 0) != SQLITE_OK)  ||
        (sqlite3_bind_int (stmt, 9, data.song_number)        != SQLITE_OK)
       ) {
        LOG_ERROR_SCT("DB", "Failed to bind for kara %s: %s", filename, sqlite3_errmsg((sqlite3 *) db));
        goto error;
    }

    SQLITE_STEP_DONE(db, stmt, error);
    status = true;
error:
    sqlite3_finalize(stmt);
    return status;
}

bool
database_update_set_available(volatile sqlite3 *db, uint64_t id)
{
    static const char *SQL_STMT = "UPDATE kara SET available = 1 WHERE id = ?;";
    sqlite3_stmt *stmt;
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, (int) id, error);
    if (sqlite3_step(stmt) != SQLITE_DONE)
        goto error;
    sqlite3_finalize(stmt);
    LOG_INFO_SCT("DB", "Kara %ld is now available", id);
    return true;
error:
    sqlite3_finalize(stmt);
    LOG_ERROR_SCT("DB", "Failed to set kara %ld available", id);
    return false;
}

bool
database_update_add(volatile sqlite3 *db, const char *kara_path, struct kara_metadata *mdt, uint64_t id, bool avail)
{
    RETURN_UNLESS(db && kara_path && mdt, "Invalid argument", false);
    static const char *SQL_STMT =
        "INSERT OR REPLACE INTO "
        "kara (song_name, source_name, category, song_type, language, file_path, is_new, author_name, author_year, song_number, id, available)"
        "SELECT ?, ?, ?, ?, ?, ?, ?, ?, strftime('%s','now'), ?, ?, ?";
    sqlite3_stmt *stmt = NULL;
    int code = SQLITE_OK, ret = false;
    time_t the_time = time(NULL);
    struct tm *the_local_time = localtime(&the_time);
    char year[10];
    code = snprintf(year, 10, "%d", the_local_time->tm_year + 1900);
    GOTO_IF(code < 0 || code >= 10, "Failed to get the year of the current date", error_no_sqlite);

    /* From here we initialize the sqlite stmt. */

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);

    if ((sqlite3_bind_text(stmt,  1, mdt->song_name,   -1, 0) != SQLITE_OK) ||
        (sqlite3_bind_text(stmt,  2, mdt->source_name, -1, 0) != SQLITE_OK) ||
        (sqlite3_bind_text(stmt,  3, mdt->category,    -1, 0) != SQLITE_OK) ||
        (sqlite3_bind_text(stmt,  4, mdt->song_type,   -1, 0) != SQLITE_OK) ||
        (sqlite3_bind_text(stmt,  5, mdt->language,    -1, 0) != SQLITE_OK) ||
        (sqlite3_bind_text(stmt,  6, kara_path,        -1, 0) != SQLITE_OK) ||
        (sqlite3_bind_int (stmt,  7, 0) != SQLITE_OK) /* TODO */            ||  /* No new kara added (TODO). */
        (sqlite3_bind_text(stmt,  8, mdt->author_name, -1, 0) != SQLITE_OK) ||
        (sqlite3_bind_int (stmt,  9, mdt->song_number)        != SQLITE_OK) ||
        (sqlite3_bind_int (stmt, 10, id)                      != SQLITE_OK) ||
        (sqlite3_bind_int (stmt, 11, avail)                   != SQLITE_OK)
       ) {
        LOG_ERROR_SCT("DB", "Failed to bind argument for kara %s: %s", kara_path, sqlite3_errmsg((sqlite3 *) db));
        goto error;
    }

    SQLITE_STEP_DONE(db, stmt, error);
    ret = true;
error:
    sqlite3_finalize(stmt);
error_no_sqlite:
    return ret;
}

bool
database_update(volatile sqlite3 *db, const char *kara_dir)
{
    DIR *d;
    struct dirent *dir;
    char path[PATH_MAX];
    memset(path, 0, PATH_MAX * sizeof(char));

    if (!(d = opendir(kara_dir))) {
        LOG_ERROR("Failed to open directory '%s': %s", kara_dir, strerror(errno));
        return false;
    }

    while (NULL != (dir = readdir(d))) {
        strncpy(path, kara_dir, PATH_MAX - 1);
        strncat(path, "/", PATH_MAX - 1);
        strncat(path, dir->d_name, PATH_MAX - 1);

        if (dir->d_type == DT_REG) {
            if (get_mtime(path) < database_get_timestamp(db))
                continue;
            database_add_kara(db, path);
            database_stamp(db);
        }

        else if (dir->d_type == DT_DIR          &&
                 strcmp(dir->d_name, ".")  != 0 &&
                 strcmp(dir->d_name, "..") != 0)
            database_update(db, path);
    }

    LOG_INFO("Passed directory '%s'", kara_dir);
    database_updated(db);
    closedir(d);
    return true;
}
