#define _POSIX_C_SOURCE 200809L

#include <common/common.h>
#include <lektor/module/module_x11.h>
#include <lektor/common.h>
#include <lektor/module/mpv.h>
#include <lektor/database.h>
#include <lektor/commands.h>

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <mpv/client.h>

#define WIDTH   400
#define HEIGHT  200

struct module_x11_window {
    /* X11 related, the master and slave window */
    Display *master_display;
    Display *child_display;
    Window master_win;
    Window child_win;

    /* X11 related, some X11 properties */
    int screen;
    XSetWindowAttributes attrs;
    Atom wm_delete;
    XColor black;
    bool is_fullscreen;

    /* Mpv related */
    mpv_handle *mpv;
    volatile int mpv_time_pos;  // Don't write it in the database       //
    volatile int mpv_duration;  // Because don't need to be persistent  //
};

/*
 * NOW: THE ONLY EXPORTED FUNCTION :NOW
 */

extern int
module_set_function(void *arg__, void *handle__)
{
    RETURN_UNLESS(arg__ && handle__, "Invalid arguments", 1);
    struct lkt_win *win = (struct lkt_win *) arg__;

    win->new            = module_x11_new;
    win->close          = module_x11_close;
    win->free           = module_x11_free;
    win->toggle_pause   = module_x11_toggle_pause;
    win->load_file      = module_x11_load_file;
    win->set_volume     = module_x11_set_volume;
    win->get_duration   = module_x11_get_duration;
    win->get_elapsed    = module_x11_get_elapsed;
    win->handle_events  = module_x11_handle_events;
    win->attach         = module_x11_attach;
    win->handle         = handle__;

    return 0;
}

/*
 *
 * NOW: PRIVATE X11 FUNCTIONS :NOW
 *
 */

static inline bool
lx11_new(struct module_x11_window *ret)
{
    ret->master_display = XOpenDisplay(NULL);
    RETURN_UNLESS(ret->master_display, "Can't open display", false);
    ret->screen = DefaultScreen(ret->master_display);
    Colormap colormap = DefaultColormap(ret->master_display, ret->screen);
    XAllocNamedColor(ret->master_display, colormap, "black", &ret->black, &ret->black);

    // Initialize container window attributes
    ret->attrs.event_mask
        = SubstructureRedirectMask                  // handle child window requests         (MapRequest)
          | SubstructureNotifyMask                  // handle child window notifications    (DestroyNotify)
          | StructureNotifyMask                     // handle container notifications       (ConfigureNotify)
          | ExposureMask                            // handle container redraw              (Expose)
          | KeyReleaseMask                          // handle release key                   (KeyRelease)
          | VisibilityChangeMask                    // handle... I don't know
          ;
    ret->attrs.do_not_propagate_mask = 1;           // do not hide any events from child window
    ret->attrs.background_pixel = ret->black.pixel; // background color
    unsigned long attrs_mask = CWEventMask          // enable attrs.event_mask
                               | NoEventMask        // enable attrs.do_not_propagate_mask
                               | CWBackPixel        // enable attrs.background_pixel
                               ;

    // Create and map container window
    ret->master_win = XCreateWindow(ret->master_display, RootWindow(ret->master_display, ret->screen),
                                    0, 0, WIDTH, HEIGHT, 1, CopyFromParent, InputOutput,
                                    CopyFromParent, attrs_mask, &ret->attrs);

    XMapWindow(ret->master_display, ret->master_win);                               // Make window visible
    XStoreName(ret->master_display, ret->master_win, "Lektor Player");              // Set window title
    ret->wm_delete = XInternAtom(ret->master_display, "WM_DELETE_WINDOW", True);    // Get WM_DELETE_WINDOW atom
    XSetWMProtocols(ret->master_display, ret->master_win, &ret->wm_delete, 1);

    ret->child_display = NULL;
    ret->child_win = 0;

    return true;
}

static inline bool
lx11_handle(struct module_x11_window *win)
{
    XEvent event;

    RETURN_UNLESS(win && win->mpv, "Invalid argument", true);

    for (;;) {

        if (!XCheckWindowEvent(win->master_display, win->master_win, win->attrs.event_mask, &event))
            return true;

        // Map child window when it requests and store its display and window id
        if (event.type == MapRequest) {
            XMapWindow(event.xmaprequest.display, event.xmaprequest.window);

            win->child_display = event.xmaprequest.display;
            win->child_win = event.xmaprequest.window;
        }

        // Propagate resize event to child window, and also resize it after MapRequest
        if (event.type == ConfigureNotify || event.type == MapRequest) {
            if (win->child_win) {
                // Get container window attributes
                XWindowAttributes attrs;
                XGetWindowAttributes(win->master_display, win->master_win, &attrs);

                // Move and resize child
                XMoveResizeWindow(win->child_display, win->child_win,
                                  2, 2, attrs.width - 6, attrs.height - 6);
            }
        }

        // Key press
        if (event.type == KeyRelease) {
            LOG_INFO_SCT("WINDOW", "Release key: 0x%x", event.xkey.keycode);
            if (event.xkey.keycode == 0x47)
                XClearWindow(win->master_display, win->master_win);
        }

        // Refresh container window
        if (event.type == Expose)
            XClearWindow(win->master_display, win->master_win);

        // Close button
        if (event.type == ClientMessage) {
            if ((long unsigned int) event.xclient.data.l[0] == win->wm_delete)
                return false;
        }

    }
    return true;
}

/*
 *
 * NOW: THE SETTED FUNCTIONS FOR THE WINDOW MODULE :NOW
 *
 */

bool
module_x11_new(struct lkt_win *const win)
{
    RETURN_UNLESS(win, "Invalid argument", false);
    struct module_x11_window *x11_win = win->window;

    if (x11_win == NULL) {
        x11_win = calloc(1, sizeof(struct module_x11_window));
        memset(x11_win, 0, sizeof(struct module_x11_window));
        RETURN_UNLESS(x11_win, "Out of memory", false);
        RETURN_UNLESS(lx11_new(x11_win), "Can't create X11 window", false);
        LOG_INFO_SCT("WINDOW", "%s", "Created the X11 window");
    }

    if (x11_win->mpv == NULL) {
        x11_win->mpv = lmpv_new(x11_win->master_win);
        RETURN_UNLESS(x11_win->mpv, "Failed to create mpv", false);
    }

    win->window = x11_win;
    return true;
}

void
module_x11_close(struct lkt_win *const win)
{
    RETURN_UNLESS(win && win->window, "Invalid arguments", NOTHING);
    struct module_x11_window *const x11_win = win->window;
    lmpv_free(&x11_win->mpv);
    x11_win->child_display = NULL;
    x11_win->child_win = 0;
}

void
module_x11_free(struct lkt_win *const win)
{
    RETURN_UNLESS(win && win->window, "Invalid arguments", NOTHING);
    struct module_x11_window *const x11_win = win->window;
    lmpv_free(&x11_win->mpv);
    XCloseDisplay(x11_win->master_display);
    free(x11_win);
    win->window = NULL;
}

void
module_x11_attach(struct lkt_win *const win, void *server)
{
    UNUSED(win, server);
}

bool
module_x11_toggle_pause(struct lkt_win *const win)
{
    RETURN_UNLESS(win && win->window, "Invalid arguments", false);
    return ! lmpv_toggle_pause(((struct module_x11_window *) win->window)->mpv);
}

bool
module_x11_load_file(struct lkt_win *const win, const char *filepath)
{
    RETURN_UNLESS(win && win->window, "Invalid arguments", false);
    struct module_x11_window *xwin = win->window;
    bool ret = ! lmpv_load_file(xwin->mpv, filepath);
    xwin->mpv_duration = 0;
    xwin->mpv_time_pos = 0;
    return ret;
}

bool
module_x11_set_volume(struct lkt_win *const win, int vol)
{
    RETURN_UNLESS(win && win->window, "Invalid arguments", false);
    return ! lmpv_set_volume(((struct module_x11_window *) win->window)->mpv, vol);
}

bool
module_x11_get_duration(struct lkt_win *const win, int *dur_sec)
{
    RETURN_UNLESS(win && win->window, "Invalid arguments", false);
    struct module_x11_window *win_x11 = win->window;
    *dur_sec = win_x11->mpv_duration;
    return true;
}

bool
module_x11_get_elapsed(struct lkt_win *const win, int *elapsed_sec)
{
    RETURN_UNLESS(win && win->window, "Invalid arguments", false);
    struct module_x11_window *win_x11 = win->window;
    *elapsed_sec = win_x11->mpv_time_pos;
    return true;
}

bool
module_x11_handle_events(struct lkt_win *const win, volatile sqlite3 *db, mpd_idle_flag *mpd_idle_events)
{
    RETURN_UNLESS(win && win->window, "Invalid arguments", false);
    struct module_x11_window *xwin = win->window;
    return lx11_handle(xwin) &&
           ! lmpv_handle(win, xwin->mpv, db, mpd_idle_events, &xwin->mpv_time_pos, &xwin->mpv_duration);
}
