#define _POSIX_C_SOURCE 200809L

#include <mthread/mthread_internal.h>
#include <sched.h>
#include <string.h>
#include <unistd.h>

#define MTHREAD_LWP 1

static mthread_virtual_processor_t virtual_processors[MTHREAD_MAX_VIRUTAL_PROCESSORS];
static mthread_list_t joined_list;

/* Is mthread initialized */
static volatile int is_mthread_init = 0;

static void
__mthread_clear_keys(mthread_t *thread)
{
    RETURN_UNLESS(thread, "Invalid argument", NOTHING);
    unsigned int i;
    struct keys_list *keys = &(*thread)->keys;

    if (!keys) {
        keys->first_avail = 0;
        return;
    }

    for (i = 0; i < keys->first_avail; ++i)
        mthread_key_delete(keys->list[i]);

    keys->first_avail = 0;
}

static inline void
mthread_list_init(mthread_list_t *list)
{
    mthread_list_t INIT = MTHREAD_LIST_INIT;
    *list = INIT;
}

static inline void
mthread_init_thread(struct mthread_s *thread)
{
    thread->detached = JOINABLE;
    thread->next     = NULL;
    thread->status   = RUNNING;
    thread->res      = NULL;
}

void
mthread_insert_first(struct mthread_s *item, mthread_list_t *list)
{
    mthread_spinlock_lock(&(list->lock));
    if (list->first == NULL) {
        item->next  = NULL;
        list->first = item;
        list->last  = item;
    } else {
        item ->next = list->first;
        list->first = item;
    }
    mthread_spinlock_unlock(&(list->lock));
}

void
mthread_insert_last(struct mthread_s *item, mthread_list_t *list)
{
    mthread_spinlock_lock(&(list->lock));
    if (list->first == NULL) {
        item->next  = NULL;
        list->first = item;
        list->last  = item;
    } else {
        item->next = NULL;
        list->last->next = item;
        list->last = item;
    }
    mthread_spinlock_unlock(&(list->lock));
}

struct mthread_s *
mthread_remove_first(mthread_list_t *list)
{
    struct mthread_s *res = NULL;
    mthread_spinlock_lock(&(list->lock));
    if (list->first != NULL) {
        res = (struct mthread_s *)list->first;
        list->first = res->next;
        if (list->first == NULL)
            list->last = NULL;
    }
    mthread_spinlock_unlock(&(list->lock));
    return res;
}

extern int
mthread_list_test(struct mthread_s *item, mthread_list_t *list)
{
    struct mthread_s *it = (struct mthread_s *) list->last;
    while (it && it != item)
        it = (struct mthread_s *) it->next;
    return it == item;
}

static inline int
mthread_mctx_set(struct mthread_s *mctx, void (*func) (void *), char *stack, size_t size, void *arg)
{
    /* fetch current context */
    RETURN_IF(getcontext(&(mctx->uc)) != 0, "Error in getcontext", 1);

    /* remove parent link */
    mctx->uc.uc_link = NULL;

    /* configure new stack */
    mctx->uc.uc_stack.ss_sp    = stack;
    mctx->uc.uc_stack.ss_size  = size;
    mctx->uc.uc_stack.ss_flags = 0;
    mctx->stack                = stack;

    /* configure startup function (with one argument) */
    makecontext(&(mctx->uc), (void (*)(void)) func, 1 + 1, arg);
    return 0;
}

static inline int
mthread_mctx_swap(struct mthread_s *cur_mctx, struct mthread_s *new_mctx)
{
    swapcontext(&(cur_mctx->uc), &(new_mctx->uc));
    return 0;
}

void
__mthread_yield(mthread_virtual_processor_t *vp)
{
    struct mthread_s *current = (struct mthread_s *)vp->current;
    struct mthread_s *next    = mthread_remove_first(&(vp->ready_list));

    if (vp->resched != NULL) {
        mthread_insert_last((struct mthread_s *)vp->resched, &(vp->ready_list));
        vp->resched = NULL;
    }

    if (current != vp->idle) {
        if ((current->status != BLOCKED) && (current->status != ZOMBIE)) {
            if (current->status == RUNNING)
                vp->resched = current;

            else
                not_implemented();
        }

        if (next == NULL)
            next = vp->idle;
    }

    /* always true at this point - except for idle thread */
    if (next != NULL) {
        if (vp->current != next) {
            vp->current = next;
            mthread_mctx_swap(current, next);
        }
    }

    vp = mthread_get_vp();

    if (vp->p != NULL) {
        mthread_spinlock_unlock(vp->p);
        vp->p = NULL;
    }
}

static void
mthread_idle_task(void *arg)
{
    mthread_virtual_processor_t *vp = arg;
    long j;
    int done = 0;

    vp->state = 1;
    while (done == 0) {
        done = 1;
        sched_yield();
        for (j = 0; j < MTHREAD_LWP; j++) {
            if (virtual_processors[j].state == 0)
                done = 0;
        }
    }
    LOG_INFO_SCT("SCHEDULER", "Virtual processor %d started", vp->rank);
    while (1)
        __mthread_yield(vp);
    not_implemented();
}

mthread_virtual_processor_t *
mthread_get_vp()
{
    return &(virtual_processors[0]);
}

int
mthread_get_vp_rank()
{
    return mthread_get_vp()->rank;
}

static inline void
mthread_init_vp(mthread_virtual_processor_t *vp, struct mthread_s *idle, struct mthread_s *current, int rank)
{
    vp->current = current;
    vp->idle    = idle;
    mthread_list_init(&(vp->ready_list));
    vp->rank    = rank;
    vp->resched = NULL;
    vp->p       = NULL;
}

void *
mthread_main(void *arg)
{
    UNUSED(arg);
    mthread_virtual_processor_t *vp = mthread_get_vp();
    while (1) {
        sched_yield();
        sleep(1);
        __mthread_yield(vp);
    }
    return NULL;
}

static inline void
mthread_init_lib(long i)
{
    struct mthread_s *current = NULL;
    char *stack               = (char *) safe_malloc(MTHREAD_DEFAULT_STACK);
    struct mthread_s *mctx    = (struct mthread_s *) safe_malloc(sizeof(struct mthread_s));
    mthread_init_thread(mctx);
    mthread_list_init(&(joined_list));

    if (i == 0) {
        current = (struct mthread_s *)safe_malloc(sizeof(struct mthread_s));
        mthread_init_thread(current);
        current->__start_routine = mthread_main;
        current->stack = NULL;
    }

    mthread_init_vp(&(virtual_processors[i]), mctx, mctx, i);
    mthread_mctx_set(mctx, mthread_idle_task, stack, MTHREAD_DEFAULT_STACK, &(virtual_processors[i]));
    if (i != 0) {
        virtual_processors[i].current = mctx;
        setcontext(&(mctx->uc));
    } else
        virtual_processors[i].current = current;
}

static void
mthread_start_thread(void *arg)
{
    struct mthread_s *mctx = (struct mthread_s *)arg;
    LOG_INFO_SCT("THREAD", "Thread %p started", arg);
    mthread_virtual_processor_t *vp = mthread_get_vp();
    __mthread_yield(vp);
    mctx->res    = mctx->__start_routine(mctx->arg);
    mctx->status = ZOMBIE;
    LOG_INFO_SCT("THREAD", "Thread %p ended (%d)", arg, vp->rank);
    if (mctx->detached) {
        LOG_INFO_SCT("THREAD", "Thread %p (vp: %d) was detached, join it", arg, vp->rank);
        mthread_join(mctx, NULL);
    }
    vp = mthread_get_vp();
    __mthread_yield(vp);
}

/* Function for handling threads.  */

static inline void
__mthread_lib_init(long i)
{
    mthread_init_lib(i);
    virtual_processors[0].state = 1;
    LOG_INFO_SCT("THREAD", "%s", "library started");
}

void
mthread_init(void)
{
    if (is_mthread_init == 0) {
        __mthread_lib_init(0);
        is_mthread_init = 1;
    }
}

/* Create a thread with given attributes ATTR (or default attributes
   if ATTR is NULL), and call function START_ROUTINE with given
   arguments ARG.  */
int
mthread_create(mthread_t *__threadp, const mthread_attr_t __attr, void *(*__start_routine) (void *), void *__arg)
{
    if (is_mthread_init == 0) {
        __mthread_lib_init(0);
        is_mthread_init = 1;
    }

    mthread_virtual_processor_t *vp = mthread_get_vp();

    struct mthread_s *mctx = mthread_remove_first(&(joined_list));
    if (mctx == NULL) {
        mctx = safe_malloc(sizeof(struct mthread_s));
        memset(mctx, 0, sizeof(struct mthread_s));
    }
    char *stack = mctx->stack ? mctx->stack : safe_malloc(MTHREAD_DEFAULT_STACK);

    mthread_init_thread(mctx);

    if (__attr) {
        if (__attr & ATTR_DETACHED)
            mctx->detached = DETACHED;
        if (__attr & ATTR_DETACHED_FREE)
            mctx->detached = DETACHED_FREE;
    }

    LOG_INFO_SCT("THREAD INIT", "Create thread %p", (void *) mctx);
    mctx->arg = __arg;
    mctx->__start_routine = __start_routine;
    mthread_mctx_set(mctx, mthread_start_thread, stack, MTHREAD_DEFAULT_STACK, mctx);
    mthread_insert_last(mctx, &(vp->ready_list));

    if (__threadp)
        *__threadp = mctx;

    return 0;
}

/* Obtain the identifier of the current thread.  */
mthread_t
mthread_self(void)
{
    return (mthread_t) mthread_get_vp()->current;
}

/* Compare two thread identifiers.  */
int
mthread_equal(mthread_t __thread1, mthread_t __thread2)
{
    return (__thread1 == __thread2);
}

/* Terminate calling thread.  */
void
mthread_exit(void *__retval)
{
    mthread_virtual_processor_t *vp = mthread_get_vp();
    struct mthread_s *mctx          = (struct mthread_s *) vp->current;
    mctx->res    = __retval;
    mctx->status = ZOMBIE;
    LOG_INFO_SCT("THREAD END", "Thread %p exited", (void *) mctx);
    __mthread_yield(vp);
}

/* Make calling thread wait for termination of the thread TH.  The
   exit status of the thread is stored in *THREAD_RETURN, if THREAD_RETURN
   is not NULL.  */
int
mthread_join(mthread_t __th, void **__thread_return)
{
    LOG_INFO_SCT("THREAD END", "Join thread %p", (void *) __th);

    while (__th->status != ZOMBIE)
        mthread_yield();

    if (__thread_return != NULL)
        *__thread_return = (void *)__th->res;

    __mthread_clear_keys(&__th);
    LOG_INFO_SCT("THREAD END", "Thread %p joined", (void *) __th);
    mthread_insert_last(__th, &(joined_list));
    return 0;
}

void
mthread_yield()
{
    mthread_virtual_processor_t *vp = mthread_get_vp();
    LOG_INFO_SCT("THREAD YIELD", "Thread %p yield", (void *) vp->current);
    __mthread_yield(vp);
}
